// Copyright (C) 2023 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:flutter/material.dart';

/// {@template theme_resolver}
/// In charge of theme negotiation and merge.
///
/// When you build a component `Component({double? radius})`.
/// You have several possibilities:
/// 1) Pass the "radius" into the constructor, `Component(radius: 12)`.
/// 2) Set up a theme extension `ComponentThemeExtension(radius: 15)`.
/// 3) Let `wyatt_ui_kit` "negotiate" and try to find a suitable style in the
/// flutter theme.
///
/// If this negotiation phase fails, then:
/// - If the value is mandatory: a hardcoded value in "wyatt_ui_kit" is chosen.
/// - If not, the style is simply not applied.
/// {@endtemplate}
abstract class ThemeResolver<S, T, E> {
  /// {@macro theme_resolver}
  const ThemeResolver();

  S? Function(
    BuildContext context,
    S defaultValue,
    T themeExtension, {
    E? extra,
  }) get computeExtensionValueFn;
  S? Function(BuildContext context, {E? extra}) get customStyleFn;

  /// Compute default value from Flutter Theme or with hardcoded values.
  S computeDefaultValue(BuildContext context, {E? extra});

  /// Compute values from the extension if found
  S? computeExtensionValue(
    BuildContext context,
    S defaultValue, {
    E? extra,
  }) {
    final themeExtension = findExtension(context);
    if (themeExtension != null) {
      return computeExtensionValueFn(context, defaultValue, themeExtension);
    }
    return defaultValue;
  }

  /// Compute custom value
  S? computeCustomValue(
    BuildContext context,
    S previousPhaseValue, {
    E? extra,
  }) {
    final customStyle = customStyleFn(context, extra: extra);
    if (customStyle != null) {
      return customStyle;
    }
    return previousPhaseValue;
  }

  T? findExtension(BuildContext context) => Theme.of(context).extension<T>();

  /// Choose most suitable style for a given context.
  S negotiate(BuildContext context, {E? extra}) {
    S style = computeDefaultValue(context, extra: extra);
    style = computeExtensionValue(context, style, extra: extra) ?? style;
    style = computeCustomValue(context, style, extra: extra) ?? style;
    return style;
  }
}
